package sketch;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Locale;

import fileIO.TextStreamWriter;
import tax.PrintTaxonomy;
import tax.TaxNode;
import tax.TaxTree;

public class SketchObject {
	
	static synchronized void setTaxtree(String taxTreeFile){
		if(taxTreeFile==null){
			taxtree=null;
			return;
		}
		if(treefile!=null){
			assert(!treefile.equals(taxTreeFile));
			if(treefile.equals(taxTreeFile)){return;}
			treefile=taxTreeFile;
		}
		taxtree=TaxTree.loadTaxTree(taxTreeFile, System.err, hashNames);
	}
	
	static int parseMode(String arg, String a, String b){
		int mode_=-1;
		if(a.equals("mode")){
			if(b.equalsIgnoreCase("single") || b.equalsIgnoreCase("onesketch")){
				mode_=ONE_SKETCH;
			}else if(b.equalsIgnoreCase("taxa") || b.equalsIgnoreCase("pertaxa")){
				mode_=PER_TAXA;
			}else if(b.equalsIgnoreCase("sequence") || b.equalsIgnoreCase("persequence")){
				mode_=PER_SEQUENCE;
			}else if(b.equalsIgnoreCase("img")){
				mode_=IMG;
			}else if(Character.isDigit(b.charAt(0))){
				mode_=Integer.parseInt(b);
			}
		}else if(a.equalsIgnoreCase("single") || a.equalsIgnoreCase("onesketch")){
			mode_=ONE_SKETCH;
		}else if(a.equalsIgnoreCase("taxa") || a.equalsIgnoreCase("pertaxa")){
			mode_=PER_TAXA;
		}else if(a.equalsIgnoreCase("sequence") || a.equalsIgnoreCase("persequence")){
			mode_=PER_SEQUENCE;
		}
		return mode_;
	}
	
	static void formatComparisonColumnwise(Comparison c, StringBuilder sb, boolean printTax){
			sb.append(String.format(Locale.ROOT, "%.2f%%\t%.2f%%\t%d\t%d\t%d\t%d\t",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex, c.taxID, c.genomeSize));
			if(taxtree==null || !printTax){
				sb.append(c.taxName==null ? '.' : c.taxName);
			}else{
				if(c.taxID>=0 && c.taxID<minFakeID){
					TaxNode tn=taxtree.getNode(c.taxID);
					if(tn!=null){
						sb.append(taxtree.toSemicolon(tn, skipNonCanonical));
					}
				}
			}
			if(printOriginalName){
				sb.append('\t');
				sb.append(c.name0==null ? "." : c.name0);
			}
			sb.append('\n');
	}
	
	static String header(){
		String s="WKID\tKID\tmatches\tcompared\ttaxID\tgSize\ttaxName"+(printOriginalName ? "\tseqName" : "");
		return s;
	}
	
	static void formatComparison(Comparison c, int format, StringBuilder sb, boolean printTax){
		if(format==2){
			formatComparisonColumnwise(c, sb, printTax);
			return;
		}
		if(format==0){
			sb.append(String.format(Locale.ROOT, "WKID %.2f%%\tKID %.2f%%\tmatches %d\tcompared %d",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex)+"\ttaxID "+c.taxID+"\tgSize "+c.genomeSize+"\t"+
					(c.taxName==null ? "." : c.taxName)+
					(printOriginalName ? "\t"+(c.name0==null ? "." : c.name0) : "")+"\n");
			if(printTax && taxtree!=null){
				if(c.taxID>=0 && c.taxID<minFakeID){
					TaxNode tn=taxtree.getNode(c.taxID);
					if(tn!=null){
						PrintTaxonomy.printTaxonomy(tn, sb, taxtree, TaxTree.DOMAIN, skipNonCanonical);
					}
				}
				sb.append('\n');
			}
		}else{
			ArrayList<TaxNode> tnl=new ArrayList<TaxNode>();
			if(taxtree!=null && c.taxID>=0 && c.taxID<minFakeID){
				TaxNode tn=taxtree.getNode(c.taxID);
				while(tn!=null && tn.pid!=tn.id && tn.level<=TaxTree.DOMAIN){
					tnl.add(tn);
					tn=taxtree.getNode(tn.pid);
				}
			}
			
			sb.append(String.format(Locale.ROOT, "WKID %.2f%%\tKID %.2f%%\tmatches %d\tcompared %d\t",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex));
			sb.append("\ttaxID ").append(c.taxID).append('\t');
			sb.append(c.taxName).append('\t');
			if(printOriginalName){sb.append(c.name0).append('\t');}
			
			for(int i=tnl.size()-1; i>=0; i--){
				TaxNode tn=tnl.get(i);
				sb.append(tn.name);
				if(i>0){sb.append(';');}
			}
			sb.append('\n');
			
			tnl.clear();
		}
	}
	
	public static TaxNode parseNodeFromHeader(String header, boolean bestEffort){
		if(taxtree==null){return null;}
		TaxNode tn;
		if(!SILVA_MODE){
			tn=taxtree.getNode(header, '|');
		}else{
			tn=taxtree.getNodeSilva(header, bestEffort);
		}
		return tn;
	}
	

	
	/*--------------------------------------------------------------*/
	/*----------------         Print Methods        ----------------*/
	/*--------------------------------------------------------------*/
	
	void writeResults(ArrayList<Comparison> al, Sketch s, int format, boolean printTax, TextStreamWriter tsw){
		tsw.print("\nResults for "+s.name()+":\n\n");
		
		StringBuilder sb=new StringBuilder();
		if(format==2){sb.append(header()).append('\n');}
		for(Comparison c : al){	
			formatComparison(c, format, sb, printTax);
			tsw.print(sb.toString());
			sb.setLength(0);
		}
	}
	
	StringBuilder toText(ArrayList<Comparison> al, Sketch s, int format, boolean printTax){
		StringBuilder sb=new StringBuilder();
		
		sb.append("\nResults for "+s.name()+":\n\n");

		if(format==2){sb.append(header()).append('\n');}
		for(Comparison c : al){	
			formatComparison(c, format, sb, printTax);
		}
		return sb;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------        Static Fields         ----------------*/
	/*--------------------------------------------------------------*/
	
	public static final int RAW=0, HEX=1, A48=2;
	public static final char[] codingArray={'R', 'H', 'A'};
	
	public static int CODING=A48;
	public static boolean delta=true;
	public static float maxGenomeFraction=0.02f;
	public static boolean amino=false;
	public static boolean SILVA_MODE=false;
	
	public static boolean printOriginalName=true;
	
	static final byte[] hexTable=new byte[128];
	static {
		Arrays.fill(hexTable, (byte)-1);
		for(int i='0'; i<='9'; i++){
			hexTable[i]=(byte)(i-'0');
		}
		for(int i='A'; i<='F'; i++){
			hexTable[i]=hexTable[i+'a'-'A']=(byte)(i-'A'+10);
		}
		hexTable['x']=hexTable['X']=hexTable['-']=hexTable['+']=0;
	}
	
	public static TaxTree taxtree=null;
	private static String treefile=null;
	static String blacklist=null;
	
	static int minFakeID=1900000000;
	static boolean hashNames=false;
	static boolean skipNonCanonical=true;
	
	static long maxReads=-1;
	static float samplerate=1.0f;
	static long sampleseed=-1L;
	
	public static final int ONE_SKETCH=1, PER_SEQUENCE=2, PER_TAXA=3, IMG=4;
	
}
